@@ -1,6 +1,7 @@
 use std::{fmt, process::Command};
 
 use crate::{
+    buffer::char_display_len,
     command::{CommandManager, CommandTokenizer},
     editor::{BufferedKeys, EditorContext, EditorFlow, KeysIterator},
     events::{KeyParseAllError, KeyParser},
@@ -186,6 +187,13 @@ pub enum MessageKind {
     Error,
 }
 
+#[derive(Default)]
+pub struct StatusBarDisplay<'status_bar, 'lines> {
+    pub prefix: &'static str,
+    pub prefix_is_line: bool,
+    pub lines: &'lines [&'status_bar str],
+}
+
 pub struct StatusBar {
     kind: MessageKind,
     message: String,
@@ -198,22 +206,102 @@ impl StatusBar {
         }
     }
 
-    pub fn message(&self) -> (MessageKind, &str) {
-        (self.kind, &self.message)
+    pub fn is_empty(&self) -> bool {
+        self.message.is_empty()
     }
 
     pub fn clear(&mut self) {
         self.message.clear();
     }
 
-    pub fn write(&mut self, kind: MessageKind) -> EditorOutputWrite {
+    pub fn write(&mut self, kind: MessageKind) -> StatusBarWriter {
         self.kind = kind;
         self.message.clear();
-        EditorOutputWrite(&mut self.message)
+        StatusBarWriter(&mut self.message)
+    }
+
+    pub(crate) fn on_before_render(&mut self) {
+        let trimmed_len = self.message.trim_end().len();
+        self.message.truncate(trimmed_len);
+
+        unsafe {
+            for b in self.message.as_mut_vec().iter_mut() {
+                if *b == b'\t' {
+                    *b = b' ';
+                }
+            }
+        }
+    }
+
+    pub(crate) fn display<'this, 'lines>(
+        &'this self,
+        available_size: (u16, u8),
+        lines: &'lines mut [&'this str],
+    ) -> StatusBarDisplay<'this, 'lines> {
+        let lines = if lines.len() > available_size.1 as _ {
+            &mut lines[..available_size.1 as usize]
+        } else {
+            lines
+        };
+
+        let prefix = match self.kind {
+            MessageKind::Info => "",
+            MessageKind::Error => "error: ",
+        };
+
+        let mut lines_len = 0;
+        let mut x = 0;
+        let mut line_start_index = 0;
+        let mut prefix_is_line = false;
+
+        if (prefix.len() + self.message.len()) < available_size.0 as _ {
+            x = prefix.len();
+        } else {
+            prefix_is_line = !prefix.is_empty();
+        }
+
+        for (i, c) in self.message.char_indices() {
+            match c {
+                '\n' => {
+                    if lines_len >= lines.len() {
+                        break;
+                    }
+
+                    lines[lines_len] = &self.message[line_start_index..i];
+                    lines_len += 1;
+                    line_start_index = i + 1;
+                }
+                c => {
+                    let c_len = char_display_len(c) as usize;
+                    x += c_len;
+                    if x >= available_size.0 as _ {
+                        x = c_len;
+
+                        if lines_len >= lines.len() {
+                            break;
+                        }
+
+                        lines[lines_len] = &self.message[line_start_index..i];
+                        lines_len += 1;
+                        line_start_index = i;
+                    }
+                }
+            }
+        }
+        if lines_len < lines.len() && line_start_index < self.message.len() {
+            lines[lines_len] = &self.message[line_start_index..];
+            lines_len += 1;
+        }
+
+        StatusBarDisplay {
+            prefix,
+            prefix_is_line,
+            lines: &lines[..lines_len],
+        }
     }
 }
-pub struct EditorOutputWrite<'a>(&'a mut String);
-impl<'a> EditorOutputWrite<'a> {
+pub struct StatusBarWriter<'a>(&'a mut String);
+impl<'a> StatusBarWriter<'a> {
     pub fn str(&mut self, message: &str) {
         self.0.push_str(message);
     }
